
/* Chrysalide - Outil d'analyse de fichiers binaires
 * coder.c - lecture automatisée des spécifications d'architecture
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Chrysalide.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "coder.h"


#include <assert.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>


#include "helpers.h"



/* -------------------------- CONSTRUCTION SELON COMMANDES -------------------------- */


/* Suivi des constructions */
struct _rented_coder
{
    const char *input;                      /* Fichier de définitions      */
    const char *arch;                       /* Architecture à traiter      */
    const char *header;                     /* En-tête pour les en-têtes   */
    const char *const_prefix;               /* Préfixe pour les opérandes  */

    pre_processor *pp;                      /* Pré-processeur avec macros  */

    char *copyright;                        /* Récupération des droits     */
    char *ins;                              /* Désignation humaine         */
    char separator;                         /* Caractère de séparation     */
    char *raw_details;                      /* Eventuels compléments bruts */
    char *details;                          /* Eventuels compléments       */

    instr_id *id;                           /* Gestionnaire d'identifiant  */
    instr_desc *desc;                       /* Gestionnaire de description */

    bool useless;                           /* Instruction non utilisée    */

    encoding_spec **specs;                  /* Définitions déjà en place   */
    size_t specs_count;                     /* Nombre de ces définitions   */
    encoding_spec *cur_spec;                /* Définition courante         */

};



/* --------------------------- REPRESENTATION D'ENCODAGES --------------------------- */


/* Détermine le nombre de bits analysés lors d'un désassemblage. */
static unsigned int get_bit_width_for_encoding_spec(const rented_coder *, const string_exch *);



/* --------------------------- GENERATIONS DE CODE SOURCE --------------------------- */


/* Ouvre un fichier principal en écriture pour y placer du code. */
static int open_instr_header_file(const rented_coder *, const output_info *, const char *, bool *);

/* Ouvre un fichier principal en écriture pour y placer du code. */
static int open_instr_code_file(const rented_coder *, const output_info *, const char *, bool *);

/* Ouvre un fichier global en écriture pour y placer du code. */
static int open_global_header_file(const rented_coder *, const output_info *, const char *, bool *);

/* Imprime dans un flux donné un commentaire de propriété. */
static void write_header_file_license(int, const output_info *, const char *, const char *);

/* Imprime dans un flux donné un commentaire de propriété. */
static void write_code_file_license(int, const output_info *, const char *, const char *);

/* Initialise le contenu utile du fichier des instructions. */
static void init_coder_opcodes_file(int, const output_info *, const char *);

/* Initialise le contenu utile d'un fichier d'instructions. */
static void init_coder_code_file(int, const char *);

/* Centralise l'impression du nom de fonction de désassemblage. */
static void write_read_function_name(int fd, const char *, const string_exch *, const char *);

/* Génère ou complète un fichier contenant le code C principal. */
static bool output_coder_raw(const rented_coder *, const output_info *, const string_exch *, const encoding_spec *, int, int);

/* Génère ou complète un fichier contenant le code C principal. */
static bool output_coder_main_raw(const rented_coder *, const output_info *, const string_exch *, int, int);

/* Génère ou complète un fichier contenant le code C principal. */
static bool output_coder_format(const rented_coder *, const output_info *, const string_exch *, const encoding_spec *, int, int);

/* Initialise le contenu utile du fichier des identifiants. */
static void init_coder_identifiers_file(int, const output_info *);

/* Initialise le contenu utile du fichier des sous-identifiants. */
static void init_coder_sub_identifiers_file(int, const output_info *);

/* Initialise le contenu utile du fichier des mots clefs. */
static void init_coder_keywords_file(int, const output_info *);

/* Initialise le contenu utile du fichier des décrochages. */
static void init_coder_hooks_file(int, const output_info *);

/* Initialise le contenu utile du fichier des descriptions. */
static void init_coder_descriptions_file(int, const output_info *);



/* ---------------------------------------------------------------------------------- */
/*                            CONSTRUCTION SELON COMMANDES                            */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : pp = préprocesseur déjà chargé à intégrer.                   *
*                                                                             *
*  Description : Débute la définition d'une fonction de désassemblage.        *
*                                                                             *
*  Retour      : Gestionnaire mis en place.                                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

rented_coder *create_coder(pre_processor *pp)
{
    rented_coder *result;                   /* Structure à renvoyer        */

    result = (rented_coder *)calloc(1, sizeof(rented_coder));

    result->pp = pp;

    result->id = create_instruction_id();
    result->desc = create_instruction_description();

    result->cur_spec = create_encoding_spec();

    result->useless = false;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Supprime le codeur de la mémoire.                            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void delete_coder(rented_coder *coder)
{
    size_t i;                               /* Boucle de parcours          */

    delete_pre_processor(coder->pp);

    if (coder->ins != NULL)
        free(coder->ins);

    if (coder->raw_details != NULL)
    {
        free(coder->raw_details);
        free(coder->details);
    }

    delete_instruction_id(coder->id);
    delete_instruction_description(coder->desc);

    for (i = 0; i < coder->specs_count; i++)
        delete_encoding_spec(coder->specs[i]);

    if (coder->specs != NULL)
        free(coder->specs);

    delete_encoding_spec(coder->cur_spec);

    free(coder);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Détermine si les propriétés de base d'un codeur sont là.     *
*                                                                             *
*  Retour      : Bilan de l'état opérationnel.                                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool do_basic_checks_with_coder(const rented_coder *coder)
{
    /*
    result = coder->type != IOT_UNDEFINED && coder->outdir != NULL;
    result &= coder->arch != NULL && coder->header != NULL;

    if (coder->type == IOT_FORMAT)
        result &= (coder->const_prefix != NULL);
    */
    return false;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Indique le fichier de définition considéré en entrée.        *
*                                                                             *
*  Retour      : Fichier de définition à interpréter.                         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

const char *get_coder_input_file(const rented_coder *coder)
{
    return coder->input;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                input = fichier de définitions à venir lire.                 *
*                                                                             *
*  Description : Spécifie le fichier de définition à prendre en entrée.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void set_coder_input_file(rented_coder *coder, const char *input)
{
    coder->input = input;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                type  = type de définition à attendre.                       *
*                                                                             *
*  Description : Spécifie le type de format à prendre en compte (E/S).        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/
/*
void set_coder_input_type(rented_coder *coder, InputOutputType type)
{
    coder->type = type;

}
*/


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                arch  = désignation pour le code de l'architecture lue.      *
*                                                                             *
*  Description : Détermine l'architecture visée par les traitements.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void set_coder_arch(rented_coder *coder, const char *arch)
{
    coder->arch = arch;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder  = gestion par la machine en remplacement de l'humain. *
*                header = base des définitions de protection d'en-têtes.      *
*                                                                             *
*  Description : Définit la base des protections des fichiers d'en-tête.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void set_coder_header_base(rented_coder *coder, const char *header)
{
    coder->header = header;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder  = gestion par la machine en remplacement de l'humain. *
*                prefix = préfixe pour les types d'opérandes.                 *
*                                                                             *
*  Description : Définit le préfixe pour les opérandes chargées par format.   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void set_coder_const_prefix(rented_coder *coder, const char *prefix)
{
    coder->const_prefix = prefix;

}



/* ---------------------------------------------------------------------------------- */
/*                               INFORMATIONS GENERALES                               */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : coder   = gestion par la machine en remplacement de l'humain.*
*                copy    = droits de copie en anglais.                        *
*                ins     = désignation humaine de l'instruction.              *
*                sep     = caractère de séparation avant les détails.         *
*                details = compléments d'informations éventuels ou NULL.      *
*                                                                             *
*  Description : Enregistre les contours d'une instruction d'assemblage.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void save_notes_for_coder(rented_coder *coder, char *copy, char *ins, char sep, const char *details)
{
    coder->copyright = copy;
    coder->ins = make_string_lower(ins);

    coder->separator = sep;

    if (details != NULL)
    {
        coder->raw_details = strdup(details);
        coder->details = make_callable(details, true);
    }
    else
    {
        coder->raw_details = strdup("");
        coder->details = strdup("");
    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Fournit la désignation nominale d'une instruction.           *
*                                                                             *
*  Retour      : Désignation nominale à libérer de la mémoire.                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

char *get_coder_nominal_name(const rented_coder *coder)
{
    char *result;                           /* Désignation à retourner     */

    if (coder->separator == '\0')
        result = strdup(coder->ins);

    else
        asprintf(&result, "%s%c%s", coder->ins, coder->separator, coder->raw_details);

    return result;

}

/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Fournit la désignation complète d'une instruction.           *
*                                                                             *
*  Retour      : Désignation complète à libérer de la mémoire.                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

char *get_coder_code_name(const rented_coder *coder)
{
    char *result;                           /* Désignation à retourner     */
    char *keyword;                          /* Mot clef appelable en code  */
    char *details;                          /* Compléments de distinction  */

    keyword = make_callable(coder->ins, false);

    if (coder->separator == '\0')
        result = keyword;

    else
    {
        details = make_callable(coder->raw_details, true);

        asprintf(&result, "%s%s", keyword, details);

        free(keyword);

        free(details);

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Fournit le gestionnaire des définitions d'identifiant.       *
*                                                                             *
*  Retour      : Structure assurant la définition d'identifiant.              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

instr_id *get_coder_instruction_id(const rented_coder *coder)
{
    return coder->id;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Fournit le gestionnaire de description d'identifiant.        *
*                                                                             *
*  Retour      : Structure assurant la description d'identifiant.             *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

instr_desc *get_coder_instruction_desc(const rented_coder *coder)
{
    return coder->desc;

}



/* ---------------------------------------------------------------------------------- */
/*                             REPRESENTATION D'ENCODAGES                             */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Fournit un lien vers les spécifications courantes.           *
*                                                                             *
*  Retour      : Spécification en cours d'édition.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

encoding_spec *get_current_encoding_spec(const rented_coder *coder)
{
    return coder->cur_spec;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder  = gestion par la machine en remplacement de l'humain. *
*                prefix = distinction principale entre les définitions.       *
*                index  = distinction secondaire entre les définitions.       *
*                                                                             *
*  Description : Enregistre une définition supplémentaire.                    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void push_encoding_spec(rented_coder *coder, char *prefix, unsigned int index)
{
    encoding_spec *spec;                    /* Définition à compléter      */

    spec = coder->cur_spec;

    if (prefix != NULL)
        define_encoding_spec_code_name(spec, prefix, index);

    coder->specs = (encoding_spec **)realloc(coder->specs, ++coder->specs_count * sizeof(encoding_spec *));
    coder->specs[coder->specs_count - 1] = spec;

    coder->cur_spec = create_encoding_spec();

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder    = gestion automatique de l'écriture de code.        *
*                enc_name = désignation du type d'encodage visé.              *
*                                                                             *
*  Description : Détermine le nombre de bits analysés lors d'un désassemblage.*
*                                                                             *
*  Retour      : Nombre de bits interprété.                                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static unsigned int get_bit_width_for_encoding_spec(const rented_coder *coder, const string_exch *enc_name)
{
    unsigned int result;                    /* Taille à retourner          */
    size_t i;                               /* Boucle de parcours          */
    encoding_spec *spec;                    /* Définition à traiter        */
    coding_bits *bits;                      /* Gestionnaire de bits        */

    result = -1;

    for (i = 0; i < coder->specs_count; i++)
    {
        spec = coder->specs[i];

        if (!has_encoding_spec_prefix(spec, enc_name->src))
            continue;

        bits = get_bits_in_encoding_spec(spec);
        result = count_coded_bits(bits);
        break;

    }

    /**
     * Rien n'a été trouvé à faire...
     * Cette situation doit normalement être écartée par l'appelant,
     * afin d'éviter de constituer des fichiers vides.
     */
    assert(result != -1);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                                                                             *
*  Description : Marque une instruction comme non utilisée.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void mark_coder_as_useless(rented_coder *coder)
{
    coder->useless = true;

}



/* ---------------------------------------------------------------------------------- */
/*                             GENERATIONS DE CODE SOURCE                             */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : coder  = gestion par la machine en remplacement de l'humain. *
*                info   = précisions quant à la génération.                   *
*                prefix = type d'encodage à répercuter sur le nom de fichier. *
*                new    = dit si l'opération a abouti à une création. [OUT]   *
*                                                                             *
*  Description : Ouvre un fichier principal en écriture pour y placer du code.*
*                                                                             *
*  Retour      : Descripteur du fichier ouvert ou -1 en cas d'échec.          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int open_instr_header_file(const rented_coder *coder, const output_info *info, const char *prefix, bool *new)
{
    int result;                             /* Descripteur à retourner     */
    char *pathname;                         /* Chemin d'accès à constituer */
    int ret;                                /* Test d'existence du fichier */
    int flags;                              /* Mode d'accès au fichier     */

    if (prefix != NULL)
        asprintf(&pathname, "%s%s_opcodes.h", info->directory, prefix);
    else
        asprintf(&pathname, "%sopcodes.h", info->directory);

    ret = access(pathname, F_OK);

    *new = (ret != 0);

    if (*new)
        flags = O_WRONLY | O_CREAT;
    else
        flags = O_WRONLY | O_APPEND;

    result = open(pathname, flags, 0644);
    if (result == -1) perror("open()");

    free(pathname);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder  = gestion par la machine en remplacement de l'humain. *
*                info   = précisions quant à la génération.                   *
*                prefix = type d'encodage à répercuter sur le nom de fichier. *
*                new    = dit si l'opération a abouti à une création. [OUT]   *
*                                                                             *
*  Description : Ouvre un fichier principal en écriture pour y placer du code.*
*                                                                             *
*  Retour      : Descripteur du fichier ouvert ou -1 en cas d'échec.          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int open_instr_code_file(const rented_coder *coder, const output_info *info, const char *prefix, bool *new)
{
    int result;                             /* Descripteur à retourner     */
    char *group;                            /* Regroupement des similarités*/
    char *sep;                              /* Eventuelle séparation       */
    char *pathname;                         /* Chemin d'accès à constituer */
    int ret;                                /* Test d'existence du fichier */
    int flags;                              /* Mode d'accès au fichier     */

    group = strdup(coder->ins);

    sep = index(group, '-');

    if (sep != NULL)
        *sep = '\0';

    if (prefix != NULL)
        asprintf(&pathname, "%s%s_%s.c", info->directory, prefix, group);
    else
        asprintf(&pathname, "%s%s.c", info->directory, group);

    free(group);

    ret = access(pathname, F_OK);

    *new = (ret != 0);

    if (*new)
        flags = O_WRONLY | O_CREAT;
    else
        flags = O_WRONLY | O_APPEND;

    result = open(pathname, flags, 0644);
    if (result == -1) perror("open()");

    free(pathname);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                name  = nom du fichier ciblé par l'opération.                *
*                new   = indique si l'opération a créé le fichier ciblé. [OUT]*
*                                                                             *
*  Description : Ouvre un fichier global en écriture pour y placer du code.   *
*                                                                             *
*  Retour      : Descripteur du fichier ouvert ou -1 en cas d'échec.          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int open_global_header_file(const rented_coder *coder, const output_info *info, const char *name, bool *new)
{
    int result;                             /* Descripteur à retourner     */
    char *pathname;                         /* Chemin d'accès à constituer */
    int ret;                                /* Test d'existence du fichier */
    int flags;                              /* Mode d'accès au fichier     */

    asprintf(&pathname, "%s%s.h", info->directory, name);

    ret = access(pathname, F_OK);

    *new = (ret != 0);

    if (*new)
        flags = O_WRONLY | O_CREAT;
    else
        flags = O_WRONLY | O_APPEND;

    result = open(pathname, flags, 0644);
    if (result == -1) perror("open()");

    free(pathname);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture mis à disposition.            *
*                info = précisions quant à la génération.                     *
*                name = nom du fichier ciblé par l'opération.                 *
*                msg  = complément d'information à faire paraître.            *
*                                                                             *
*  Description : Imprime dans un flux donné un commentaire de propriété.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void write_header_file_license(int fd, const output_info *info, const char *name, const char *msg)
{
    time_t seconds;                         /* Temps écoulé depuis T0      */
    struct tm cur_date;                     /* Informations sur la date    */

    time(&seconds);
    localtime_r(&seconds, &cur_date);

    dprintf(fd, "\n");

    dprintf(fd, "/* Chrysalide - Outil d'analyse de fichiers binaires\n");
    dprintf(fd, " * %s.h - %s %s\n", name, msg, info->arch);
    dprintf(fd, " *\n");
    dprintf(fd, " * Copyright (C) %d Cyrille Bagard\n", 1900 + cur_date.tm_year);
    dprintf(fd, " *\n");
    dprintf(fd, " *  This file is part of Chrysalide.\n");
    dprintf(fd, " *\n");
    dprintf(fd, " *  Chrysalide is free software; you can redistribute it and/or modify\n");
    dprintf(fd, " *  it under the terms of the GNU General Public License as published by\n");
    dprintf(fd, " *  the Free Software Foundation; either version 3 of the License, or\n");
    dprintf(fd, " *  (at your option) any later version.\n");
    dprintf(fd, " *\n");
    dprintf(fd, " *  Chrysalide is distributed in the hope that it will be useful,\n");
    dprintf(fd, " *  but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    dprintf(fd, " *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    dprintf(fd, " *  GNU General Public License for more details.\n");
    dprintf(fd, " *\n");
    dprintf(fd, " *  You should have received a copy of the GNU General Public License\n");
    dprintf(fd, " *  along with Chrysalide.  If not, see <http://www.gnu.org/licenses/>.\n");
    dprintf(fd, " */\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd        = flux ouvert en écriture mis à disposition.       *
*                info      = précisions quant à la génération.                *
*                name      = nom du fichier ciblé par l'opération.            *
*                copyright = droits d'auteur à faire valoir.                  *
*                                                                             *
*  Description : Imprime dans un flux donné un commentaire de propriété.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void write_code_file_license(int fd, const output_info *info, const char *name, const char *copyright)
{
    dprintf(fd, "\n");

    dprintf(fd, "/* Chrysalide - Outil d'analyse de fichiers binaires\n");
    dprintf(fd, " * %s.c - traduction d'instructions %s\n", name, info->arch);
    dprintf(fd, " *\n");
    dprintf(fd, " * %s\n", copyright);
    dprintf(fd, " *\n");
    dprintf(fd, " *  This file is part of Chrysalide.\n");
    dprintf(fd, " *\n");
    dprintf(fd, " *  Chrysalide is free software; you can redistribute it and/or modify\n");
    dprintf(fd, " *  it under the terms of the GNU General Public License as published by\n");
    dprintf(fd, " *  the Free Software Foundation; either version 3 of the License, or\n");
    dprintf(fd, " *  (at your option) any later version.\n");
    dprintf(fd, " *\n");
    dprintf(fd, " *  Chrysalide is distributed in the hope that it will be useful,\n");
    dprintf(fd, " *  but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    dprintf(fd, " *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    dprintf(fd, " *  GNU General Public License for more details.\n");
    dprintf(fd, " *\n");
    dprintf(fd, " *  You should have received a copy of the GNU General Public License\n");
    dprintf(fd, " *  along with Chrysalide.  If not, see <http://www.gnu.org/licenses/>.\n");
    dprintf(fd, " */\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd     = flux ouvert en écriture mis à disposition.          *
*                info   = précisions quant à la génération.                   *
*                prefix = préfixe lié à une sous-branche de l'architecture.   *
*                                                                             *
*  Description : Initialise le contenu utile du fichier des instructions.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_opcodes_file(int fd, const output_info *info, const char *prefix)
{
    char *sub;                              /* Sous-partie à intégrer      */

    if (prefix != NULL)
    {
        sub = strdup(prefix);
        make_string_upper(sub);

        dprintf(fd, "#ifndef _%s_%s_OPCODES_H\n", info->guard, sub);
        dprintf(fd, "#define _%s_%s_OPCODES_H\n", info->guard, sub);

        free(sub);

    }

    else
    {
        dprintf(fd, "#ifndef _%s_OPCODES_H\n", info->guard);
        dprintf(fd, "#define _%s_OPCODES_H\n", info->guard);
    }

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "##INCLUDES##\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd     = flux ouvert en écriture mis à disposition.          *
*                prefix = type d'encodage à répercuter sur un nom de fichier. *
*                                                                             *
*  Description : Initialise le contenu utile d'un fichier d'instructions.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_code_file(int fd, const char *prefix)
{
    if (prefix != NULL)
        dprintf(fd, "#include \"%s_opcodes.h\"\n", prefix);
    else
        dprintf(fd, "#include \"opcodes.h\"\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "##INCLUDES##\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                                                                             *
*  Description : Génère ou complète un fichier contenant le code C principal. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool output_coder_body(const rented_coder *coder, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    size_t i;                               /* Boucle de parcours #1       */
    const string_exch *enc_name;            /* Type d'encodage visé        */
    size_t j;                               /* Boucle de parcours #2       */
    int header_fd;                          /* Fichier de déclarations     */
    char *file;                             /* Nom de fichier final        */
    bool header_new;                        /* Note une création d'entête  */
    int code_fd;                            /* Fichier de définitions      */
    bool code_new;                          /* Note une création de code   */

    result = true;

    for (i = 0; i < count_encodings(coder->pp) && result; i++)
    {
        enc_name = find_encoding(coder->pp, i);

        for (j = 0; j < coder->specs_count; j++)
        {
            /* On s'assure qu'il existe bien une version pour l'encodage visé... */
            if (!has_encoding_spec_prefix(coder->specs[j], enc_name->src))
                continue;

            header_fd = open_instr_header_file(coder, info, enc_name->dest, &header_new);
            if (header_fd == -1)
            {
                result = false;
                goto ocb_exit;
            }

            if (header_new)
            {
                if (enc_name->dest == NULL)
                    file = strdup("opcodes");
                else
                    asprintf(&file, "%s_opcodes", enc_name->dest);

                write_header_file_license(header_fd, info, file, "prototypes pour la traduction d'instructions");

                free(file);

                init_coder_opcodes_file(header_fd, info, enc_name->dest);

            }

            code_fd = open_instr_code_file(coder, info, enc_name->dest, &code_new);
            if (code_fd == -1)
            {
                result = false;
                goto ocb_exit;
            }

            if (code_new)
            {
                if (enc_name->dest == NULL)
                    file = strdup(coder->ins);
                else
                    asprintf(&file, "%s_%s", enc_name->dest, coder->ins);

                write_code_file_license(code_fd, info, file, coder->copyright);

                free(file);

                init_coder_code_file(code_fd, enc_name->dest);

            }
            else
                dprintf(code_fd, "\n");

            switch (info->type)
            {
                case IOT_UNDEFINED:
                    assert(false);
                    result = false;
                    break;

                case IOT_RAW:
                    result = output_coder_raw(coder, info, enc_name, coder->specs[j], header_fd, code_fd);
                    break;

                case IOT_FORMAT:
                    assert(j == 0);
                    assert(enc_name->dest == NULL);
                    result = output_coder_format(coder, info, enc_name, coder->specs[j], header_fd, code_fd);
                    break;

            }

        }

        /* La suite ne concerne que les formats bruts aboutis... */
        if (!result) break;
        if (info->type != IOT_RAW) continue;

        for (j = 0; j < coder->specs_count; j++)
        {
            /* On s'assure de retrouver une version de l'encodage visé juste avant... */
            if (!has_encoding_spec_prefix(coder->specs[j], enc_name->src))
                continue;

            header_fd = open_instr_header_file(coder, info, enc_name->dest, &header_new);
            if (header_fd == -1)
            {
                result = false;
                goto ocb_exit;
            }

            assert(!header_new);

            code_fd = open_instr_code_file(coder, info, enc_name->dest, &code_new);
            if (code_fd == -1)
            {
                result = false;
                goto ocb_exit;
            }

            assert(!code_new);

            result = output_coder_main_raw(coder, info, enc_name, header_fd, code_fd);

            break;

        }

    }

 ocb_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture.                              *
*                arch = architecture en cours de traitement.                  *
*                sub  = encodage choisi comme sous-ensemble d'architecture.   *
*                name = désignation complète d'une instruction.               *
*                                                                             *
*  Description : Centralise l'impression du nom de fonction de désassemblage. *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void write_read_function_name(int fd, const char *arch, const string_exch *sub, const char *name)
{
    if (sub->dest == NULL)
        dprintf(fd, "%s_read_instr_%s", arch, name);
    else
        dprintf(fd, "%s_read_%s_instr_%s", arch, sub->dest, name);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder    = gestion automatique de l'écriture de code.        *
*                info     = précisions quant à la génération.                 *
*                enc_name = désignation du type d'encodage visé.              *
*                encoding = sélection de l'encodage à traiter.                *
*                hfd      = flux ouvert en écriture pour les déclarations.    *
*                cfd      = flux ouvert en écriture pour les définitions.     *
*                                                                             *
*  Description : Génère ou complète un fichier contenant le code C principal. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool output_coder_raw(const rented_coder *coder, const output_info *info, const string_exch *enc_name, const encoding_spec *encoding, int hfd, int cfd)
{
    bool result;                            /* Bilan à retourner           */
    char *arch;                             /* Architecture à traiter      */
    char *name;                             /* Désignation à manipuler     */
    char *prefix;                           /* Préfixe employé en suffixe  */
    coding_bits *bits;                      /* Gestionnaire de bits        */
    unsigned int wide;                      /* Taille des mots             */
    size_t maxlen;                          /* Taille à compléter          */
    char *fullname;                         /* Désignation complète        */

    arch = strdup(info->arch_cn);
    make_string_lower(arch);

    name = get_coder_code_name(coder);

    prefix = build_encoding_spec_prefix(encoding);

    bits = get_bits_in_encoding_spec(encoding);
    wide = count_coded_bits(bits);

    /* Désassemblage : déclaration */

    if (0 /* TODO : export seulement */)
    {
        dprintf(hfd, "/* Décode une forme d'instruction de type '%s'. */\n", coder->ins);

        dprintf(hfd, "GArchInstruction *");
        write_read_function_name(hfd, arch, enc_name, name);
        dprintf(hfd, "_%s", prefix);

        dprintf(hfd, "(");
        dprintf(hfd, "uint%u_t raw", wide);
        dprintf(hfd, ");\n");

        dprintf(hfd, "\n");

    }

    /* Désassemblage : définitions */

    dprintf(cfd, "/******************************************************************************\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Paramètres  : raw = données brutes à analyser.                             *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Description : Décode une forme d'instruction de type '%s'.", coder->ins);

    maxlen = 20 - strlen(coder->ins);

    if (maxlen < 20)
        dprintf(cfd, "%*s\n", (int)maxlen, "*");
    else
        dprintf(cfd, "*\n");

    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Retour      : Instruction mise en place ou NULL en cas d'échec.            *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Remarques   : -                                                            *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "******************************************************************************/\n");

    dprintf(cfd, "\n");

    if (1 /* TODO : si pas exportée */)
        dprintf(cfd, "static ");

    dprintf(cfd, "GArchInstruction *");
    write_read_function_name(cfd, arch, enc_name, name);
    dprintf(cfd, "_%s", prefix);

    dprintf(cfd, "(");
    dprintf(cfd, "uint%u_t raw", wide);
    dprintf(cfd, ")");

    dprintf(cfd, "\n");
    dprintf(cfd, "{");
    dprintf(cfd, "\n");

    make_string_upper(name);

    asprintf(&fullname, "%s_%s", info->id_prefix, name);

    result = write_encoding_spec_raw_disass(encoding, cfd, arch, fullname, coder->pp);

    free(fullname);

    dprintf(cfd, "}\n");
    dprintf(cfd, "\n");

    /* Conclusion */

    free(prefix);

    free(name);

    free(arch);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder    = gestion automatique de l'écriture de code.        *
*                info     = précisions quant à la génération.                 *
*                enc_name = désignation du type d'encodage visé.              *
*                hfd      = flux ouvert en écriture pour les déclarations.    *
*                cfd      = flux ouvert en écriture pour les définitions.     *
*                                                                             *
*  Description : Génère ou complète un fichier contenant le code C principal. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool output_coder_main_raw(const rented_coder *coder, const output_info *info, const string_exch *enc_name, int hfd, int cfd)
{
    bool result;                            /* Bilan à retourner           */
    char *arch;                             /* Architecture à traiter      */
    char *name;                             /* Désignation à manipuler     */
    unsigned int wide;                      /* Taille des mots             */
    size_t maxlen;                          /* Taille à compléter          */
    bool first;                             /* Note un premier appel       */
    size_t i;                               /* Boucle de parcours          */
    char *prefix;                           /* Préfixe employé en suffixe  */

    result = false;

    arch = strdup(info->arch_cn);
    make_string_lower(arch);

    name = get_coder_code_name(coder);

    wide = get_bit_width_for_encoding_spec(coder, enc_name);

    /* Désassemblage : déclaration */

    dprintf(hfd, "/* Décode une instruction de type '%s'. */\n", coder->ins);

    dprintf(hfd, "GArchInstruction *");
    write_read_function_name(hfd, arch, enc_name, name);

    dprintf(hfd, "(");
    dprintf(hfd, "uint%u_t raw", wide);
    dprintf(hfd, ");\n");

    dprintf(hfd, "\n");

    /* Désassemblage : définitions */

    dprintf(cfd, "/******************************************************************************\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Paramètres  : raw = données brutes à analyser.                             *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Description : Décode une instruction de type '%s'.", coder->ins);

    maxlen = 28 - strlen(coder->ins);

    if (maxlen < 28)
        dprintf(cfd, "%*s\n", (int)maxlen, "*");
    else
        dprintf(cfd, "*\n");

    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Retour      : Instruction mise en place ou NULL en cas d'échec.            *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Remarques   : -                                                            *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "******************************************************************************/\n");

    dprintf(cfd, "\n");

    dprintf(cfd, "GArchInstruction *");
    write_read_function_name(cfd, arch, enc_name, name);

    dprintf(cfd, "(");
    dprintf(cfd, "uint%u_t raw", wide);
    dprintf(cfd, ")");

    dprintf(cfd, "\n");
    dprintf(cfd, "{");
    dprintf(cfd, "\n");

    dprintf(cfd, "\tGArchInstruction *result;               /* Instruction créée à renvoyer*/\n");
    dprintf(cfd, "\n");

    first = true;

    for (i = 0; i < coder->specs_count; i++)
    {
        if (!has_encoding_spec_prefix(coder->specs[i], enc_name->src))
            continue;

        result = true;

        prefix = build_encoding_spec_prefix(coder->specs[i]);

        if (first)
        {
            dprintf(cfd, "\tresult = ");
            write_read_function_name(cfd, arch, enc_name, name);
            dprintf(cfd, "_%s(raw);\n", prefix);

            dprintf(cfd, "\n");

            first = false;

        }
        else
        {
            dprintf(cfd, "\tif (result == NULL)\n");

            dprintf(cfd, "\t\tresult = ");
            write_read_function_name(cfd, arch, enc_name, name);
            dprintf(cfd, "_%s(raw);\n", prefix);

            dprintf(cfd, "\n");

        }

        free(prefix);

    }

    dprintf(cfd, "\treturn result;\n");
    dprintf(cfd, "\n");

    dprintf(cfd, "}\n");
    dprintf(cfd, "\n");

    /* Conclusion */

    free(name);

    free(arch);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder    = gestion automatique de l'écriture de code.        *
*                info     = précisions quant à la génération.                 *
*                enc_name = désignation du type d'encodage visé.              *
*                encoding = sélection de l'encodage à traiter.                *
*                hfd      = flux ouvert en écriture pour les déclarations.    *
*                cfd      = flux ouvert en écriture pour les définitions.     *
*                                                                             *
*  Description : Génère ou complète un fichier contenant le code C principal. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool output_coder_format(const rented_coder *coder, const output_info *info, const string_exch *enc_name, const encoding_spec *encoding, int hfd, int cfd)
{
    bool result;                            /* Bilan à retourner           */
    char *arch;                             /* Architecture à traiter      */
    char *name;                             /* Désignation à manipuler     */
    size_t maxlen;                          /* Taille à compléter          */
    char *fullname;                         /* Désignation complète        */

    arch = strdup(info->arch_cn);
    make_string_lower(arch);

    name = get_coder_code_name(coder);

    /* Désassemblage : déclaration */

    dprintf(hfd, "/* Décode une instruction de type '%s'. */\n", coder->ins);

    dprintf(hfd, "GArchInstruction *");
    write_read_function_name(hfd, arch, enc_name, name);

    dprintf(hfd, "(");
    dprintf(hfd, "const GArchProcessor *, GProcContext *, const GBinContent *, ");
    dprintf(hfd, "vmpa2t *, GExeFormat *");
    dprintf(hfd, ");\n");

    dprintf(hfd, "\n");

    /* Désassemblage : définitions */

    dprintf(cfd, "/******************************************************************************\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Paramètres  : proc    = processeur de l'architecture d'exécution.          *\n");
    dprintf(cfd, "*                ctx     = contexte associé à la phase de désassemblage.      *\n");
    dprintf(cfd, "*                content = flux de données à analyser.                        *\n");
    dprintf(cfd, "*                pos     = position courante dans ce flux. [OUT]              *\n");
    dprintf(cfd, "*                format  = format du fichier contenant le code.               *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Description : Décode une instruction de type '%s'.", coder->ins);

    maxlen = 28 - strlen(coder->ins);

    if (maxlen < 28)
        dprintf(cfd, "%*s\n", (int)maxlen, "*");
    else
        dprintf(cfd, "*\n");

    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Retour      : Instruction mise en place ou NULL en cas d'échec.            *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "*  Remarques   : -                                                            *\n");
    dprintf(cfd, "*                                                                             *\n");
    dprintf(cfd, "******************************************************************************/\n");

    dprintf(cfd, "\n");

    dprintf(cfd, "GArchInstruction *");
    write_read_function_name(cfd, arch, enc_name, name);

    dprintf(cfd, "(");
    dprintf(cfd, "const GArchProcessor *proc, GProcContext *ctx, const GBinContent *content, ");
    dprintf(cfd, "vmpa2t *pos, GExeFormat *format");
    dprintf(cfd, ")");

    dprintf(cfd, "\n");
    dprintf(cfd, "{");
    dprintf(cfd, "\n");

    make_string_upper(name);

    asprintf(&fullname, "%s_%s", info->id_prefix, name);

    result = write_encoding_spec_format_disass(encoding, cfd, arch, fullname, info->fmt_prefix);

    free(fullname);

    dprintf(cfd, "}\n");
    dprintf(cfd, "\n");

    /* Conclusion */

    free(name);

    free(arch);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pathname = chemin d'accès au fichier à traiter.              *
*                info     = précisions quant à la génération.                 *
*                                                                             *
*  Description : Finalise le contenu utile du fichier des instructions.       *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool fini_coder_opcodes_file(const char *pathname, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    int fd;                                 /* Flux ouvert en écriture     */
    char *temp;                             /* Zone de travail temporaire  */
    char *base;                             /* Identification de fichier   */

    result = false;

    fd = open(pathname, O_WRONLY | O_APPEND, 0644);
    if (fd == -1)
    {
        perror("open()");
        goto fcif_exit;
    }

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    temp = strdup(pathname);
    base = basename(temp);

    base[strlen(base) - 2] = '\0';

    make_string_upper(base);

    dprintf(fd, "#endif  /* _%s_%s_H */\n", info->guard, base);

    free(temp);

    result = true;

 fcif_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture mis à disposition.            *
*                info = précisions quant à la génération.                     *
*                                                                             *
*  Description : Initialise le contenu utile du fichier des identifiants.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_identifiers_file(int fd, const output_info *info)
{
    dprintf(fd, "#ifndef _%s_IDENTIFIERS_H\n", info->guard);
    dprintf(fd, "#define _%s_IDENTIFIERS_H\n", info->guard);

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "/* Enumération de tous les opcodes */\n");
    dprintf(fd, "typedef enum _%sOpcodes\n", info->arch_cn);
    dprintf(fd, "{\n");

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                                                                             *
*  Description : Génère ou complète un fichier constituant les identifiants.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool output_coder_identifier(const rented_coder *coder, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    bool created;                           /* Note une création           */
    int fd;                                 /* Flux ouvert en écriture     */
    char *name;                             /* Désignation à manipuler     */
    instr_id *id;                           /* Gestionnaire d'identifiant  */
    unsigned int iid;                       /* Identifiant unique attribué */
    char *constant;                         /* Définition d'une constante  */
    char *comment;                          /* Contenu du commentaire      */

    result = false;

    /* Ouverture de la destination */

    fd = open_global_header_file(coder, info, "identifiers", &created);
    if (fd == -1) goto oci_exit;

    if (created)
    {
        write_header_file_license(fd, info, "identifiers", "définition des identifiants uniques pour");
        init_coder_identifiers_file(fd, info);
    }

    /* Constitution de la constante */

    name = get_coder_code_name(coder);
    make_string_upper(name);

    asprintf(&constant, "%s_%s,", info->id_prefix, name);

    free(name);

    /* Définition du commentaire */

    name = get_coder_nominal_name(coder);

    id = get_coder_instruction_id(coder);
    iid = get_instruction_id_value(id);

    asprintf(&comment, "%s (0x%0*x)", name, info->id_len, iid);

    free(name);

    /* Impression de la ligne */

    dprintf(fd, "    %-40s/* %-28s*/\n", constant, comment);

    free(constant);
    free(comment);

    result = true;

 oci_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pathname = chemin d'accès au fichier à traiter.              *
*                info     = précisions quant à la génération.                 *
*                                                                             *
*  Description : Finalise le contenu utile du fichier des identifiants.       *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool fini_coder_identifiers_file(const char *pathname, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    int fd;                                 /* Flux ouvert en écriture     */

    result = false;

    fd = open(pathname, O_WRONLY | O_APPEND, 0644);
    if (fd == -1)
    {
        perror("open()");
        goto fcif_exit;
    }

    dprintf(fd, "\n");
    dprintf(fd, "    %s_COUNT\n", info->id_prefix);
    dprintf(fd, "\n");

    dprintf(fd, "} %sOpcodes;\n", info->arch_cn);

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "#endif  /* _%s_IDENTIFIERS_H */\n", info->guard);

    result = true;

 fcif_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture mis à disposition.            *
*                info = précisions quant à la génération.                     *
*                                                                             *
*  Description : Initialise le contenu utile du fichier des sous-identifiants.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_sub_identifiers_file(int fd, const output_info *info)
{
    dprintf(fd, "#ifndef _%s_SUBIDENTIFIERS_H\n", info->guard);
    dprintf(fd, "#define _%s_SUBIDENTIFIERS_H\n", info->guard);

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "/* Enumération de tous les opcodes */\n");
    dprintf(fd, "typedef enum _%sSyntax\n", info->arch_cn);
    dprintf(fd, "{\n");

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                                                                             *
*  Description : Génère ou complète un fichier créant les sous-identifiants.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool output_coder_sub_identifier(const rented_coder *coder, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    bool created;                           /* Note une création           */
    int fd;                                 /* Flux ouvert en écriture     */
    char *name;                             /* Désignation à manipuler     */
    char *fullname;                         /* Désignation complète        */
    size_t i;                               /* Boucle de parcours          */

    result = false;

    /* Ouverture de la destination */

    fd = open_global_header_file(coder, info, "subidentifiers", &created);
    if (fd == -1) goto ocsi_exit;

    if (created)
    {
        write_header_file_license(fd, info, "subidentifiers", "définition des sous-identifiants uniques pour");
        init_coder_sub_identifiers_file(fd, info);
    }

    /* Impression des sous-identifiants */

    name = get_coder_code_name(coder);
    make_string_upper(name);

    asprintf(&fullname, "%s_%s", info->id_prefix, name);

    free(name);

    result = true;

    for (i = 0; i < coder->specs_count && result; i++)
        result = write_encoding_spec_subid(coder->specs[i], fd, fullname);

    free(fullname);

    result = true;

 ocsi_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pathname = chemin d'accès au fichier à traiter.              *
*                info     = précisions quant à la génération.                 *
*                                                                             *
*  Description : Finalise le contenu utile du fichier des sous-identifiants.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool fini_coder_sub_identifiers_file(const char *pathname, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    int fd;                                 /* Flux ouvert en écriture     */

    result = false;

    fd = open(pathname, O_WRONLY | O_APPEND, 0644);
    if (fd == -1)
    {
        perror("open()");
        goto fcif_exit;
    }

    dprintf(fd, "\n");
    dprintf(fd, "    %s_ENC_COUNT\n", info->id_prefix);
    dprintf(fd, "\n");

    dprintf(fd, "} %sSyntax;\n", info->arch_cn);

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "#endif  /* _%s_SUBIDENTIFIERS_H */\n", info->guard);

    result = true;

 fcif_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture mis à disposition.            *
*                info = précisions quant à la génération.                     *
*                                                                             *
*  Description : Initialise le contenu utile du fichier des mots clefs.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_keywords_file(int fd, const output_info *info)
{
    char *larch;                            /* Architecture en minuscules  */

    dprintf(fd, "#ifndef _%s_KEYWORDS_H\n", info->guard);
    dprintf(fd, "#define _%s_KEYWORDS_H\n", info->guard);

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    switch (info->type)
    {
        case IOT_UNDEFINED:
            assert(false);
            break;

        case IOT_RAW:
            dprintf(fd, "#include \"subidentifiers.h\"\n");
            break;

        case IOT_FORMAT:
            dprintf(fd, "#include \"identifiers.h\"\n");
            break;

    }

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    larch = strdup(info->arch_cn);
    make_string_lower(larch);

    dprintf(fd, "/* Enumération de tous les mots clefs */\n");
    dprintf(fd, "static char *_%s_keywords[%s_%sCOUNT] = {\n",
            larch, info->id_prefix, info->type == IOT_RAW ? "ENC_" : "");
    dprintf(fd, "\n");

    free(larch);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                                                                             *
*  Description : Génère ou complète un fichier constituant les mots clefs.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool output_coder_keyword(const rented_coder *coder, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    bool created;                           /* Note une création           */
    int fd;                                 /* Flux ouvert en écriture     */
    char *name;                             /* Désignation à manipuler     */
    char *fullname;                         /* Désignation complète        */
    size_t i;                               /* Boucle de parcours          */

    result = false;

    /* S'il n'y a pas lieu de traiter l'instruction */

    if (coder->useless)
    {
        result = true;
        goto ock_exit;
    }

    /* Ouverture de la destination */

    fd = open_global_header_file(coder, info, "keywords", &created);
    if (fd == -1) goto ock_exit;

    if (created)
    {
        write_header_file_license(fd, info, "keywords", "définition des mots clefs des instructions");
        init_coder_keywords_file(fd, info);
    }

    /* Lancement des impressions */

    name = get_coder_code_name(coder);
    make_string_upper(name);

    asprintf(&fullname, "%s_%s", info->id_prefix, name);

    free(name);

    result = true;

    for (i = 0; i < coder->specs_count && result; i++)
        switch (info->type)
        {
            case IOT_UNDEFINED:
                assert(false);
                result = false;
                break;

            case IOT_RAW:
                result = write_encoding_spec_keywords(coder->specs[i], fd, fullname);
                break;

            case IOT_FORMAT:
                assert(i == 0);

                /* Impression de la colonne */

                dprintf(fd, "\t[%s] = ", fullname);

                /* Impression du mot clef */

                name = get_coder_nominal_name(coder);

                dprintf(fd, "\"%s\",\n", name);

                free(name);

                break;

        }

    free(fullname);

 ock_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pathname = chemin d'accès au fichier à traiter.              *
*                info     = précisions quant à la génération.                 *
*                                                                             *
*  Description : Finalise le contenu utile du fichier des mots clefs.         *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool fini_coder_keywords_file(const char *pathname, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    int fd;                                 /* Flux ouvert en écriture     */

    result = false;

    fd = open(pathname, O_WRONLY | O_APPEND, 0644);
    if (fd == -1)
    {
        perror("open()");
        goto fckf_exit;
    }

    dprintf(fd, "\n");
    dprintf(fd, "};\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "#endif  /* _%s_KEYWORDS_H */\n", info->guard);

    result = true;

 fckf_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture mis à disposition.            *
*                info = précisions quant à la génération.                     *
*                                                                             *
*  Description : Initialise le contenu utile du fichier des décrochages.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_hooks_file(int fd, const output_info *info)
{
    char *larch;                            /* Architecture en minuscules  */

    dprintf(fd, "#ifndef _%s_HOOKS_H\n", info->guard);
    dprintf(fd, "#define _%s_HOOKS_H\n", info->guard);

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    switch (info->type)
    {
        case IOT_UNDEFINED:
            assert(false);
            break;

        case IOT_RAW:
            dprintf(fd, "#include \"subidentifiers.h\"\n");
            break;

        case IOT_FORMAT:
            dprintf(fd, "#include \"identifiers.h\"\n");
            break;

    }

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "##INCLUDES##\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    larch = strdup(info->arch_cn);
    make_string_lower(larch);

    dprintf(fd, "/* Définitions des décrochages pour l'établissement d'instructions */\n");
    dprintf(fd, "static const instr_hook_fc _%s_hooks[%s_%sCOUNT][IPH_COUNT] = {\n",
            larch, info->id_prefix, info->type == IOT_RAW ? "ENC_" : "");
    dprintf(fd, "\n");

    free(larch);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                                                                             *
*  Description : Génère ou complète un fichier constituant les décrochages.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool output_coder_hooks(const rented_coder *coder, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    bool created;                           /* Note une création           */
    int fd;                                 /* Flux ouvert en écriture     */
    char *name;                             /* Désignation à manipuler     */
    char *fullname;                         /* Désignation complète        */
    size_t i;                               /* Boucle de parcours          */

    result = false;

    /* S'il n'y a pas lieu de traiter l'instruction */

    if (coder->useless)
    {
        result = true;
        goto och_exit;
    }

    /* Ouverture de la destination */

    fd = open_global_header_file(coder, info, "hooks", &created);
    if (fd == -1) goto och_exit;

    if (created)
    {
        write_header_file_license(fd, info, "hooks", "définition des décrochages pour instructions");
        init_coder_hooks_file(fd, info);
    }

    /* Lancement des impressions */

    name = get_coder_code_name(coder);
    make_string_upper(name);

    asprintf(&fullname, "%s_%s", info->id_prefix, name);

    free(name);

    result = true;

    for (i = 0; i < coder->specs_count && result; i++)
        switch (info->type)
        {
            case IOT_UNDEFINED:
                assert(false);
                result = false;
                break;

            case IOT_RAW:
                result = write_encoding_spec_hooks(coder->specs[i], fd, fullname, true);
                break;

            case IOT_FORMAT:
                assert(i == 0);
                result = write_encoding_spec_hooks(coder->specs[i], fd, fullname, false);
                break;

        }

    free(fullname);

 och_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pathname = chemin d'accès au fichier à traiter.              *
*                info     = précisions quant à la génération.                 *
*                                                                             *
*  Description : Finalise le contenu utile du fichier des décrochages.        *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool fini_coder_hooks_file(const char *pathname, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    int fd;                                 /* Flux ouvert en écriture     */

    result = false;

    fd = open(pathname, O_WRONLY | O_APPEND, 0644);
    if (fd == -1)
    {
        perror("open()");
        goto fchf_exit;
    }

    dprintf(fd, "\n");
    dprintf(fd, "};\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "#endif  /* _%s_HOOKS_H */\n", info->guard);

    result = true;

 fchf_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : fd   = flux ouvert en écriture mis à disposition.            *
*                info = précisions quant à la génération.                     *
*                                                                             *
*  Description : Initialise le contenu utile du fichier des descriptions.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void init_coder_descriptions_file(int fd, const output_info *info)
{
    char *larch;                            /* Architecture en minuscules  */

    dprintf(fd, "#ifndef _%s_DESCRIPTIONS_H\n", info->guard);
    dprintf(fd, "#define _%s_DESCRIPTIONS_H\n", info->guard);

    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "#include \"identifiers.h\"\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    larch = strdup(info->arch_cn);
    make_string_lower(larch);

    dprintf(fd, "/* Enumération de tous les mots clefs */\n");
    dprintf(fd, "static char *_%s_descriptions[%s_COUNT] = {\n", larch, info->id_prefix);
    dprintf(fd, "\n");

    free(larch);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : coder = gestion par la machine en remplacement de l'humain.  *
*                info  = précisions quant à la génération.                    *
*                                                                             *
*  Description : Génère ou complète un fichier constituant les descriptions.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool output_coder_description(const rented_coder *coder, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    bool created;                           /* Note une création           */
    int fd;                                 /* Flux ouvert en écriture     */
    char *name;                             /* Désignation à manipuler     */

    result = false;

    /* S'il n'y a pas lieu de traiter l'instruction */

    if (coder->useless)
    {
        result = true;
        goto ocd_exit;
    }

    /* Ouverture de la destination */

    fd = open_global_header_file(coder, info, "descriptions", &created);
    if (fd == -1) goto ocd_exit;

    if (created)
    {
        write_header_file_license(fd, info, "descriptions", "définition des descriptions des instructions");
        init_coder_descriptions_file(fd, info);
    }

    /* Impression de la colonne */

    name = get_coder_code_name(coder);
    make_string_upper(name);

    dprintf(fd, "\t[%s_%s] = ", info->id_prefix, name);

    free(name);

    /* Impression du mot clef */

    name = get_coder_nominal_name(coder);

    dprintf(fd, "\"");

    write_instruction_description(coder->desc, fd);

    dprintf(fd, "\",\n");

    free(name);

    result = true;

 ocd_exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pathname = chemin d'accès au fichier à traiter.              *
*                info     = précisions quant à la génération.                 *
*                                                                             *
*  Description : Finalise le contenu utile du fichier des descriptions.       *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool fini_coder_descriptions_file(const char *pathname, const output_info *info)
{
    bool result;                            /* Bilan à retourner           */
    int fd;                                 /* Flux ouvert en écriture     */

    result = false;

    fd = open(pathname, O_WRONLY | O_APPEND, 0644);
    if (fd == -1)
    {
        perror("open()");
        goto fckf_exit;
    }

    dprintf(fd, "\n");
    dprintf(fd, "};\n");

    dprintf(fd, "\n");
    dprintf(fd, "\n");
    dprintf(fd, "\n");

    dprintf(fd, "#endif  /* _%s_DESCRIPTIONS_H */\n", info->guard);

    result = true;

 fckf_exit:

    return result;

}
